{-# LANGUAGE FunctionalDependencies #-}

module Tahoe.Download.Internal.Capability where

import Control.Monad.IO.Class
import Data.Bifunctor (Bifunctor (..))
import Data.Binary (decodeOrFail)
import qualified Data.ByteString.Lazy as LB
import qualified Data.Set as Set
import qualified Tahoe.CHK
import qualified Tahoe.CHK.Capability as CHK
import qualified Tahoe.CHK.Encrypt
import Tahoe.CHK.Server
import qualified Tahoe.CHK.Share
import Tahoe.CHK.Types
import Tahoe.Download.Internal.Client
import qualified Tahoe.SDMF as SDMF
import qualified Tahoe.SDMF.Internal.Keys as SDMF.Keys

{- | Represent the kind of share to operate on.  This forms a closed world of
 share types.  It might eventually be interesting to make an open world
 variation instead.
-}
data WhichShare
    = CHK {unWhichCHK :: Tahoe.CHK.Share.Share}
    | SDMF {unWhichSDMF :: SDMF.Share}

-- | A capability which confers the ability to locate and verify some stored data.
class Verifiable v where
    -- | Ask a storage server which share numbers related to this capability it
    -- is holding.  This is an unverified result and the storage server could
    -- present incorrect information.  Even if it correctly reports that it
    -- holds a share, it could decline to give it out when asked.
    getShareNumbers :: MonadIO m => v -> StorageServer -> m (Set.Set ShareNum)

    -- | Get the encoding parameters used for the shares of this capability.
    -- The information is presented as a tuple of (required, total).
    getRequiredTotal :: MonadIO m => v -> StorageServer -> m (Maybe (Int, Int))

    -- | Get the location information for shares of this capability.
    getStorageIndex :: v -> StorageIndex

class (Verifiable v) => Readable r v | r -> v where
    -- | Attentuate the capability.
    getVerifiable :: r -> v

    -- | Interpret the required number of shares to recover the plaintext.
    --
    -- Note: might want to split the two functions below out of decodeShare
    --
    -- shareToCipherText :: r -> [(Int, WhichShare)] -> LB.ByteString
    --
    -- cipherTextToPlainText :: r -> LB.ByteString -> LB.ByteString
    decodeShare :: MonadIO m => r -> [(Int, WhichShare)] -> m (Either DownloadError LB.ByteString)

instance Verifiable CHK.Verifier where
    getShareNumbers v s = liftIO $ storageServerGetBuckets s (CHK.storageIndex v)
    getStorageIndex CHK.Verifier{storageIndex} = storageIndex

    -- CHK is pure, we don't have to ask the StorageServer
    getRequiredTotal CHK.Verifier{required, total} _ = pure $ pure (fromIntegral required, fromIntegral total)

{- | A capability which confers the ability to interpret some stored data to
 recover the original plaintext.  Additionally, it can be attentuated to a
 Verifiable.
-}
instance Readable CHK.Reader CHK.Verifier where
    getVerifiable = CHK.verifier
    decodeShare r shareList = do
        cipherText <- liftIO $ Tahoe.CHK.decode r (second unWhichCHK <$> shareList)
        case cipherText of
            Nothing -> pure $ Left ShareDecodingFailed
            Just ct ->
                pure . Right $ Tahoe.CHK.Encrypt.decrypt (CHK.readKey r) ct

instance Verifiable SDMF.Verifier where
    getShareNumbers v s = liftIO $ storageServerGetBuckets s (SDMF.Keys.unStorageIndex $ SDMF.verifierStorageIndex v)
    getStorageIndex = SDMF.Keys.unStorageIndex . SDMF.verifierStorageIndex
    getRequiredTotal SDMF.Verifier{..} ss = do
        shareBytes <- liftIO $ storageServerRead ss (SDMF.Keys.unStorageIndex verifierStorageIndex) 0
        case decodeOrFail (LB.fromStrict shareBytes) of
            Left _ -> pure Nothing
            Right (_, _, sh) -> pure $ pure (fromIntegral $ SDMF.shareRequiredShares sh, fromIntegral $ SDMF.shareTotalShares sh)

instance Readable SDMF.Reader SDMF.Verifier where
    getVerifiable = SDMF.readerVerifier
    decodeShare r shareList = do
        cipherText <- Right <$> liftIO (SDMF.decode r (bimap fromIntegral unWhichSDMF <$> shareList))
        case cipherText of
            Left _ -> pure $ Left ShareDecodingFailed
            Right ct -> do
                print' ("Got some ciphertext: " <> show ct)
                print' ("Decrypting with iv: " <> show iv)
                pure . Right $ SDMF.decrypt readKey iv ct
              where
                readKey = SDMF.readerReadKey r
                iv = SDMF.shareIV (unWhichSDMF . snd . head $ shareList)

print' :: MonadIO m => String -> m ()
-- print' = liftIO . putStrLn
print' = const $ pure ()
